"""handle relationship between django model and graphene type. """

from functools import lru_cache
import typing

import django.db.models as djm


if typing.TYPE_CHECKING:
    import django.contrib.contenttypes.models as ctm

REGISTRY: typing.Dict[typing.Type[djm.Model], str] = {}


@lru_cache()
def get_models(v: str) -> typing.List[typing.Type[djm.Model]]:
    """Get models for typename.

    Args:
        v (str): typename

    Returns:
        typing.List[djm.Model]: list of models that registered for this typename.
    """

    ret = []
    for model, typename in REGISTRY.items():
        if typename == v:
            ret.append(model)
    return ret


@lru_cache()
def get_typename(model: typing.Type[djm.Model]) -> str:
    """Get typename for model, support inheritance.

    Args:
        model (typing.Type[djm.Model]): Model.

    Raises:
        ValueError: When model is not registered.

    Returns:
        str: Typename for this model.
    """

    if model in REGISTRY:
        return REGISTRY[model]

    for k, v in REGISTRY.items():
        if issubclass(model, k):
            return v

    raise ValueError(
        f'Model is not registed or inheritant a register model: {repr(model)}')


def get_content_type(typename: str) -> 'ctm.ContentType':
    """Get django.contrib.contenttype for typename.

    Args:
        typename (str): Graphql typename.

    Raises:
        ValueError: When not exact one matched model for given typename.

    Returns:
        ctm.ContentType: ContentType object for given typename.
    """
    # pylint: disable=import-outside-toplevel

    import django.contrib.contenttypes.models as ctm
    models = get_models(typename)
    if len(models) != 1:
        raise ValueError(
            f"Can not determinate model from typename: typename={typename}")
    return ctm.ContentType.objects.get_for_model(models[0])



"""Queryset optimization.  """

import logging
import typing

import django.db.models as djm
import graphql
import graphql.language.ast as ast_
import phrases_case

if typing.TYPE_CHECKING:
    class OptimizationOption(typing.TypedDict):
        """
        Optimization option dict.
        See :doc:`/optimize` for more information.
        """

        only: typing.Dict[typing.Optional[str], typing.List[str]]
        select: typing.Dict[typing.Optional[str], typing.List[str]]
        prefetch: typing.Dict[typing.Optional[str], typing.List[str]]
        related: typing.Dict[str, str]

    class Optimization(typing.TypedDict):
        """Optimization computation result dict.  """

        only: typing.List[str]
        select: typing.List[str]
        prefetch: typing.List[str]


LOGGER = logging.getLogger(__name__)
OPTIMIZATION_OPTIONS: typing.Dict[str, dict] = {}


def get_optimization_option(typename: str) -> 'OptimizationOption':
    """Get optimization options from typename.

    Args:
        typename (str): Graphql typename.

    Returns:
        OptimizationOption: Options.
    """

    ret = OPTIMIZATION_OPTIONS.get(
        typename,
        {},
    )
    ret.setdefault('only', {})  # type: ignore
    ret.setdefault('select', {})  # type: ignore
    ret.setdefault('prefetch', {})  # type: ignore
    ret.setdefault('related', {})  # type: ignore
    return ret  # type: ignore


def _get_inner_type(return_type):
    if not hasattr(return_type, 'of_type'):
        return return_type
    return _get_inner_type(return_type.of_type)


def _get_model_field(model: djm.Model, lookup: str) -> djm.Field:
    _model = model
    field = None
    for i in lookup.split('__'):
        field = _model._meta.get_field(i)
        _model = field.related_model
    assert field is not None
    return field


def _get_default_only_lookups(
        fieldname: str, model: djm.Model, related_query_name: str) -> typing.List[str]:
    field = None
    for lookup in (
            _format_related_name(related_query_name, fieldname),
            _format_related_name(related_query_name,
                                 phrases_case.snake(fieldname)),
    ):
        try:
            field = _get_model_field(model, lookup)
        except djm.FieldDoesNotExist:
            continue

    if field is None or not hasattr(field, 'attname'):
        return []

    return [field.attname]


def _get_selection(ast: ast_.Node, fragments, is_recursive=True) -> typing.Iterator[ast_.Field]:
    if not is_recursive and isinstance(ast, ast_.Field):
        yield ast
        return
    if isinstance(ast, (ast_.Field, ast_.FragmentDefinition, ast_.InlineFragment)):
        for i in ast.selection_set and ast.selection_set.selections or []:
            yield from _get_selection(i, fragments, is_recursive=False)
    elif isinstance(ast, ast_.FragmentSpread):
        yield from _get_selection(fragments[ast.name.value], fragments)
    else:
        raise ValueError(f'Unknown ast type: {ast}')


def _format_related_name(related_query_name, name):
    if related_query_name is 'self':
        return name
    return f'{related_query_name}__{name}'


def _get_ast_optimization(ast, return_type, fragments, model, related_query_name='self') -> 'Optimization':

    inner_type = _get_inner_type(return_type)

    opt = get_optimization_option(inner_type.name)
    ret: Optimization = {
        'only': opt['only'].get(None, []),
        'select': opt['select'].get(None, []),
        'prefetch': opt['prefetch'].get(None, []),
    }

    for sub_ast in _get_selection(ast, fragments):
        fieldname = sub_ast.name.value
        ret['only'].extend(opt['only'].get(
            fieldname) or _get_default_only_lookups(fieldname, model, related_query_name))
        ret['select'].extend(opt['select'].get(fieldname, []))
        ret['prefetch'].extend(opt['prefetch'].get(fieldname, []))

        _related_query_name = opt['related'].get(fieldname)
        if not _related_query_name:
            continue
        _optimization = _get_ast_optimization(
            sub_ast,
            inner_type.fields[fieldname].type,
            fragments, model,
            _related_query_name
        )
        ret['only'].extend(_optimization['only'])
        ret['select'].extend(_optimization['select'])
        ret['prefetch'].extend(_optimization['prefetch'])

    ret['only'] = [_format_related_name(
        related_query_name, i) for i in ret['only']]
    ret['select'] = [_format_related_name(
        related_query_name, i) for i in ret['select']]
    ret['prefetch'] = [_format_related_name(
        related_query_name, i) for i in ret['prefetch']]
    return ret


def _get_ast_and_return_type(
        info: graphql.execution.ResolveInfo,
        path: typing.Optional[typing.List[str]]
) -> typing.Tuple[graphql.GraphQLField, typing.Union[
    graphql.GraphQLList,
    graphql.GraphQLObjectType,
    graphql.GraphQLScalarType
]]:
    ret = (info.field_asts[0], info.return_type)
    for fieldname in path or []:
        ret = (
            next(
                i.name.value for i in _get_selection(ret[0], info.fragments)
                if i.name.value == fieldname
            ),
            ret[1].fields[fieldname]
        )
    return ret


def optimize(
        queryset: djm.QuerySet,
        info: graphql.ResolveInfo,
        path: typing.Optional[typing.List[str]] = None
) -> djm.QuerySet:
    """Optimization queryset with resolve info and global optimization options.

    Args:
        info (graphql.ResolveInfo): Resolve info.
        queryset (djm.QuerySet): Queryset to optimize.
        path (typing.Optional[typing.List[str]]): Field path. defaults to None.
            None means root field.

    Returns:
        djm.QuerySet: optimized queryset.
    """

    ast, return_type = _get_ast_and_return_type(info, path)
    optimization = _get_ast_optimization(
        ast, return_type, info.fragments, queryset.model)
    LOGGER.debug("Optimization queryset: optimization=%s, model=%s",
                 optimization, queryset.model)
    qs = queryset
    if optimization['select']:
        qs = qs.select_related(*optimization['select'])
    if optimization['prefetch']:
        qs = qs.prefetch_related(*optimization['prefetch'])
    qs = qs.only(*optimization['only'], *optimization['select'])
    return qs



"""Relay compatible connection resolver.  """
#pylint: disable=invalid-name

import re
import typing

import django.db.models as djm

from graphene_resolver.connection import (
    build_schema as _build_schema, get_type as _get_type,
    resolve as _resolve, resolver, _get_node_name, REGISTRY as _REGISTRY)
import graphql
import lazy_object_proxy as lazy

from . import queryset as qs_

build_schema = _build_schema
REGISTRY = _REGISTRY


def set_optimization_default(
        node: typing.Union[resolver.Resolver, str, typing.Any],
        *,
        edge_name: str = None
) -> None:
    """Set optimization default for connection.

    Args:
        node (typing.Union[resolver.Resolver, str, typing.Any]):
            Connection resolver or name or schema.
    """

    name = _get_node_name(node)
    edge_name = edge_name or f"{re.sub('Connection$', '', name)}Edge"

    opt = qs_.OPTIMIZATION_OPTIONS.setdefault(name, {})
    opt.setdefault('related', {'nodes': 'self', 'edges': 'self', })

    opt = qs_.OPTIMIZATION_OPTIONS.setdefault(edge_name, {})
    opt.setdefault('related', {'node': 'self'})


def get_type(
        node: typing.Union[resolver.Resolver, str, typing.Any],
        *,
        name: str = None,
) -> resolver.Resolver:
    """Get connection resolver from registry, and set default optimization options.
    one will be created with `build_schema` if not found in registry.

    Args:
        node (typing.Union[resolver.Resolver, str, typing.Any]): Node resolver or schema.
        name (str, optional): Override default connection name,
            required when node name is not defined.

    Returns:
        resolver.Resolver: Created connection resolver, same name will returns same resolver.
    """

    ret = _get_type(node, name=name)
    set_optimization_default(ret)
    return ret


def resolve(
        iterable,
        **kwargs,
) -> dict:
    """Resolve iterable to connection

    Args:
        iterable (typign.Iterable): value

    Returns:
        dict: Connection data.
    """

    if isinstance(iterable, djm.Manager):
        iterable = iterable.all()

    if isinstance(iterable, djm.QuerySet):
        _len = lazy.Proxy(iterable.count)
    else:
        _len = len(iterable)

    return _resolve(iterable, _len, **kwargs)


def optimized_resolve(
        info: graphql.ResolveInfo,
        queryset: djm.QuerySet,
        **kwargs,
) -> dict:
    """Resolve django queryset base on query selection.

    Args:
        info (graphql.ResolveInfo): Resolve info.
        queryset (djm.QuerySet): Queryset to resolve.

    Returns:
        dict: Connection resolve result.
    """

    qs = qs_.optimize(queryset.all(), info)
    return resolve(qs, **kwargs)



"""Utility.  """

import typing
from dataclasses import dataclass

import graphene


@dataclass
class ID:
    value: str
    type: str

    @classmethod
    def parse(cls, v: str) -> 'ID':
        """Parse graphene global id

        Args:
            v (str): value to parse

        Returns:
            ID: Parse result
        """
        try:

            type_, id_ = graphene.Node.from_global_id(v)
        except (TypeError, ValueError) as ex:
            raise ValueError(f'Invalid id: value={v}') from ex
        return cls(
            value=id_,
            type=type_
        )

    def validate_type(
            self,
            expected: typing.Union[str, typing.Tuple[str, ...]]
    ) -> 'ID':
        """Validate if id match expected type.

        Args:
            expected (typing.Union[str, typing.Tuple[str, ...]]): type name to match.

        Raises:
            ValueError: Type not match

        Returns:
            ID: self, for function chain.
        """

        expected_types = expected
        if not isinstance(expected_types, tuple):
            expected_types = (expected_types,)
        if self.type not in expected_types:
            raise ValueError(
                f'Unexpected id type: expected={expected}, actual={self.type}.')
        return self


def convert_id(v, validate_type=None):
    """Convert global id values to local db id.

    Args:
        v: value(s) to convert.
        validate_type (optional): same as `ID.validate_type` args 1. Defaults to None.

    Returns:
        Converted values.
    """
    if not v:
        return v

    values = v
    is_list = not isinstance(v, str) and isinstance(v, typing.Iterable)
    if not is_list:
        values = [values]
    id_list = [ID.parse(i) for i in values]
    if validate_type is not None:
        _ = [i.validate_type(validate_type) for i in id_list]
    ret = [i.value for i in id_list]
    if not is_list:
        ret = ret[0]
    return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphene_django_tools', 'graphene_django_tools.dataloader']

package_data = \
{'': ['*']}

install_requires = \
['graphene-resolver>=0.1.8,<0.2.0',
 'graphene>=2.1,<3.0',
 'isodate>=0.6,<0.7',
 'lazy-object-proxy>=1.4,<2.0']

setup_kwargs = {
    'name': 'graphene-django-tools',
    'version': '0.20.1',
    'description': 'Tools for use [`graphene-django`](https://github.com/graphql-python/graphene-django)',
    'long_description': '# Graphene django tools\n\n[![build status](https://github.com/NateScarlet/graphene-django-tools/workflows/Python%20package/badge.svg)](https://github.com/NateScarlet/graphene-django-tools/actions)\n[![version](https://img.shields.io/pypi/v/graphene-django-tools)](https://pypi.org/project/graphene-django-tools/)\n![python version](https://img.shields.io/pypi/pyversions/graphene-django-tools)\n![django version](https://img.shields.io/pypi/djversions/graphene-django-tools)\n![wheel](https://img.shields.io/pypi/wheel/graphene-django-tools)\n![maintenance](https://img.shields.io/maintenance/yes/2020)\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-yellow.svg)](https://conventionalcommits.org)\n\nTools for use [`graphene`](https://github.com/graphql-python/graphene) with django.\nUse a explicit schema definition approach that different from `graphene-django`.\n\nDocumentation is placed in [docs folder](./docs).\n\n## Install\n\n`pip install graphene-django-tools`\n\n## Features\n\n- django integration for [graphene-resolver](https://github.com/NateScarlet/graphene-resolver).\n- optimize queryset with django `only`,`selected_related`,`prefetch_related` to only select fields that used in query.\n- data loader graphene middleware.\n\n## Development\n\ntest: `make test`\n',
    'author': 'NateScarlet',
    'author_email': 'NateScarlet@Gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NateScarlet/graphene-django-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
